/* cmd1.c */

/* Author:
 *	Steve Kirkendall
 *	1500 SW Park #326
 *	Portland OR, 97201
 *	kirkenda@cs.pdx.edu
 */


/* This file contains some of the EX commands - mostly ones that deal with
 * files, options, etc. -- anything except text.
 */

#include "config.h"
#include "ctype.h"
#include "vi.h"
#include "regexp.h"

/*ARGSUSED*/
void cmd_mark(frommark, tomark, cmd, bang, extra)
	MARK	frommark;
	MARK	tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	/* validate the name of the mark */
	if (*extra == '"')
	{
		extra++;
	}
	/* valid mark names are lowercase ascii characters */
	if (!isascii(*extra) || !islower(*extra) || extra[1])
	{
		msg("Invalid mark name");
		return;
	}

	mark[*extra - 'a'] = tomark;
}

/*ARGSUSED*/
void cmd_write(frommark, tomark, cmd, bang, extra)
	MARK	frommark;
	MARK	tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	int		fd;
	int		append;	/* boolean: write in "append" mode? */
	REG long	l;
	REG char	*scan;
	REG int		i;

	/* if writing to a filter, then let filter() handle it */
	if (*extra == '!')
	{
		msg("UnderB does not provide filter");
		return;
	}

	/* if all lines are to be written, use tmpsave() since it is faster.
	 * Exception: if cmd_write was called by the filter() function
	 * (i.e., cmd == CMD_BANG) then do it the slow way anyway, so we
	 * can avoid clobbering prevorig.
	 */
	if (frommark == MARK_FIRST && tomark == MARK_LAST && cmd == CMD_WRITE)
	{
		tmpsave(extra, bang);
		return;
	}

	/* see if we're going to do this in append mode or not */
	append = FALSE;
	if (extra[0] == '>' && extra[1] == '>')
	{
		extra += 2;
		append = TRUE;
	}

	/* either the file must not exist, or we must have a ! or be appending */
	if (*extra && access(extra, 0) == 0 && !bang && !append)
	{
		msg("File already exists - Use :w! to overwrite");
		return;
	}

	/* else do it line-by-line, like cmd_print() */
	if (append)
	{
#ifdef O_APPEND
		fd = open(extra, O_WRONLY|O_APPEND);
#else
		fd = open(extra, O_WRONLY);
		if (fd >= 0)
		{
			lseek(fd, 0L, 2);
		}
#endif
	}
	else
	{
		fd = -1; /* so we know the file isn't open yet */
	}

	if (fd < 0)
	{
		fd = creat(extra, FILEPERMS);
		if (fd < 0)
		{
			msg("Can't write to \"%s\"", extra);
			return;
		}
	}
	for (l = markline(frommark); l <= markline(tomark); l++)
	{
		/* get the next line */
		scan = fetchline(l);
		i = strlen(scan);
		scan[i++] = '\n';

		/* print the line */
		if (twrite(fd, scan, i) < i)
		{
			msg("Write failed");
			break;
		}
	}
	rptlines = markline(tomark) - markline(frommark) + 1;
	rptlabel = "written";
	close(fd);
}	

/*ARGSUSED*/
void cmd_global(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;	/* rest of the command line */
{
	char	*cmdptr;	/* the command from the command line */
	char	cmdln[100];	/* copy of the command from the command line */
	char	*line;		/* a line from the file */
	long	l;		/* used as a counter to move through lines */
	long	lqty;		/* quantity of lines to be scanned */
	long	nchanged;	/* number of lines changed */
	regexp	*re;		/* the compiled search expression */

	/* can't nest global commands */
	if (doingglobal)
	{
		msg("Can't nest global commands.");
		rptlines = -1L;
		return;
	}

	/* ":g! ..." is the same as ":v ..." */
	if (bang)
	{
		cmd = CMD_VGLOBAL;
	}

	/* make sure we got a search pattern */
	if (!*extra)
	{
		msg("Usage: %c /regular expression/ command", cmd == CMD_GLOBAL ? 'g' : 'v');
		return;
	}

	/* parse & compile the search pattern */
	cmdptr = parseptrn(extra);
	re = regcomp(extra + 1);
	if (!re)
	{
		/* regcomp found & described an error */
		return;
	}

	/* for each line in the range */
	doingglobal = TRUE;
	ChangeText
	{
		/* NOTE: we have to go through the lines in a forward order,
		 * otherwise "g/re/p" would look funny.  *BUT* for "g/re/d"
		 * to work, simply adding 1 to the line# on each loop won't
		 * work.  The solution: count lines relative to the end of
		 * the file.  Think about it.
		 */
		for (l = nlines - markline(frommark),
			lqty = markline(tomark) - markline(frommark) + 1L,
			nchanged = 0L;
		     lqty > 0 && nlines - l >= 0 && nchanged >= 0L;
		     l--, lqty--)
		{
			/* fetch the line */
			line = fetchline(nlines - l);

			/* if it contains the search pattern... */
			if ((!regexec(re, line, 1)) == (cmd != CMD_GLOBAL))
			{
				/* move the cursor to that line */
				cursor = MARK_AT_LINE(nlines - l);

				/* do the ex command (without mucking up
				 * the original copy of the command line)
				 */
				strcpy(cmdln, cmdptr);
				rptlines = 0L;
				doexcmd(cmdln, '\\');
				nchanged += rptlines;
			}
		}
	}
	doingglobal = FALSE;

	/* free the regexp */
	_free_(re);

	/* Reporting...*/
	rptlines = nchanged;
}


/*ARGSUSED*/
void cmd_file(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	if (cmd == CMD_FILE)
	{
		msg("\"%s\" %s%s line %ld of %ld [%ld%%]",
			*origname ? origname : "[NO FILE]",
			tstflag(file, MODIFIED) ? "[MODIFIED]" : "",
			tstflag(file, READONLY) ? "[READONLY]" : "",
			markline(frommark),
			nlines,
			markline(frommark) * 100 / nlines);
	}
	else
	{
		msg("%ld", markline(frommark));
	}
}


/* This code is also used for rewind -- GB */

/*ARGSUSED*/
void cmd_next(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	int	i, j;
	char	*scan;

	/* if extra stuff given, use ":args" to define a new args list */
	if (cmd == CMD_NEXT && extra && *extra)
	{
		cmd_args(frommark, tomark, cmd, bang, extra);
	}

	/* move to the next arg */
	if (cmd == CMD_NEXT)
	{
		i = argno + 1;
	}
	else if (cmd == CMD_PREVIOUS)
	{
		i = argno - 1;
	}
	else /* cmd == CMD_REWIND */
	{
		i = 0;
	}	
	if (i < 0 || i >= nargs)
	{
		msg("No %sfiles to edit", cmd == CMD_REWIND ? "" : "more ");
		return;
	}

	/* find & isolate the name of the file to edit */
	for (j = i, scan = args; j > 0; j--)
	{
		while(*scan++)
		{
		}
	}

	/* switch to the next file */
	if (tmpabort(bang))
	{
		tmpstart(scan);
		argno = i;
	}
	else
	{
		msg("Use :%s! to abort changes, or w to save changes",
			cmd == CMD_NEXT ? "next" :
			cmd == CMD_PREVIOUS ? "previous" :
					"rewind");
	}
}

/* also called for :wq -- always writes back in this case */
/* also called for :q -- never writes back in that case */
/*ARGSUSED*/
void cmd_xit(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	static long	whenwarned;	/* when the user was last warned of extra files */
	int		oldflag;

	/* Handle :wq! command */
	if (cmd == CMD_WQUIT && bang && tstflag(file, MODIFIED)) {
		if (!tmpsave(extra, bang)) {
			msg("Could not save file");
			return;
		}
		mode = MODE_QUIT;
		return;
	}

	/* Unless the command is ":q", save the file if it has been modified */
	if (cmd != CMD_QUIT
	 && (cmd == CMD_WQUIT || tstflag(file, MODIFIED)) 
	 && !tmpsave((char *)0, FALSE) && !bang)
	{
		msg("Could not save file -- use quit! to abort changes, or w filename");
		return;
	}

	/* If there are more files to edit, then warn user */
	if (argno >= 0 && argno + 1 < nargs	/* more args */
	 && whenwarned != changes		/* user not already warned */
	 && (!bang || cmd != CMD_QUIT))		/* command not ":q!" */
	{
		msg("More files to edit -- Use \":n\" to go to next file");
		whenwarned = changes;
		return;
	}

	/* Discard the temp file.  Note that we should already have saved the
	 * the file, unless the command is ":q", so the only way that tmpabort
	 * could fail would be if you did a ":q" on a modified file.
	 */
	oldflag = *o_autowrite;
	*o_autowrite = FALSE;
	if (tmpabort(bang))
	{
		mode = MODE_QUIT;
	}
	else
	{
		msg("Use q! to abort changes, or wq to save changes");
	}
	*o_autowrite = oldflag;
}


/*ARGSUSED*/
void cmd_args(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	char	*scan;
	int	col;
	int	arg;
	int	scrolled = FALSE;
	int	width;

	/* if no extra names given, or just current name, then report the args
	 * we have now.
	 */
	if (!extra || !*extra)
	{
		/* empty args list? */
		if (nargs == 1 && !*args)
		{
			return;
		}

		/* list the arguments */
		for (scan = args, col = arg = 0;
		     arg < nargs;
		     scan += width + 1, col += width, arg++)
		{
			width = strlen(scan);
			if (col + width >= COLS - 4)
			{
				addch('\n');
				col = 0;
				scrolled = TRUE;
			}
			else if (col > 0)
			{
				addch(' ');
				col++;
			}
			if (arg == argno)
			{
				addch('[');
				addstr(scan);
				addch(']');
				col += 2;
			}
			else
			{
				addstr(scan);
			}
		}

		/* write a trailing newline */
		if ((mode == MODE_EX || mode == MODE_COLON || scrolled) && col)
		{
			addch('\n');
		}
		exrefresh();	
	}
	else /* new args list given */
	{
		for (scan = args, nargs = 1; *extra; )
		{
			if (isspace(*extra))
			{
				*scan++ = '\0';
				while (isspace(*extra))
				{
					extra++;
				}
				if (*extra)
				{
					nargs++;
				}
			}
			else if ((unsigned char)(*extra) == SPACEHOLDER)
			{
				*scan++ = ' ';
				extra++;
			}
			else
			{
				*scan++ = *extra++;
			}
		}
		*scan = '\0';

		/* reset argno to before the first, so :next will go to first */
		argno = -1;

		if (nargs != 1)
		{
                        msg("%d files to edit", nargs);
		}
	}
}


/*ARGSUSED*/
void cmd_cd(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
#ifndef CRUNCH
	/* if current file is modified, and no '!' was given, then error */
	if (tstflag(file, MODIFIED) && !bang)
	{
		msg("File modified; use \"cd! %s\" to switch anyway", extra);
	}
#endif

	/* default directory name is $HOME */
	if (!*extra)
	{
		extra = gethome((char *)0);
		if (!extra)
		{
			msg("environment variable $HOME not set");
			return;
		}
	}

	/* go to the directory */
	if (chdir(extra) < 0)
	{
		perror(extra);
	}
}


/*ARGSUSED*/
void cmd_map(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	char	*mapto;
	char	*build, *scan;

	/* "map" with no extra will dump the map table contents */
	if (!*extra)
	{
		{
			dumpkey(bang ? WHEN_VIINP|WHEN_VIREP : WHEN_VICMD, FALSE);
		}
	}
	else
	{
		/* "extra" is key to map, followed by what it maps to */

		/* handle quoting inside the "raw" string */
		for (build = mapto = extra;
		     *mapto && (
				*mapto != ' ' && *mapto != '\t');
		     *build++ = *mapto++)
		{
			if (*mapto == ctrl('V') && mapto[1])
			{
				mapto++;
			}
		}

		/* skip whitespace, and mark the end of the "raw" string */
		while ((*mapto == ' ' || *mapto == '\t'))
		{
			*mapto++ = '\0';
		}
		*build = '\0';

		/* strip ^Vs from the "cooked" string */
		for (scan = build = mapto; *scan; *build++ = *scan++)
		{
			if (*scan == ctrl('V') && scan[1])
			{
				scan++;
			}
		}
		*build = '\0';

		{
			mapkey(extra, mapto, bang ? WHEN_VIINP|WHEN_VIREP : WHEN_VICMD, (char *)0);
		}
	}
}


/*ARGSUSED*/
void cmd_set(frommark, tomark, cmd, bang, extra)
	MARK	frommark, tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	if (!*extra)
	{
		dumpopts(FALSE);/* "FALSE" means "don't dump all" - only set */
	}
	else if (!strcmp(extra, "all"))
	{
		dumpopts(TRUE);	/* "TRUE" means "dump all" - even unset vars */
	}
	else
	{
		setopts(extra);

		/* That option may have affected the appearence of text */
		changes++;
	}
}

/* describe this version of the program */
/*ARGSUSED*/
void cmd_version(frommark, tomark, cmd, bang, extra)
	MARK	frommark;
	MARK	tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
#ifdef HVERSION
	msg("%s", HVERSION);
#endif
	msg("%s", VERSION);
#ifdef CREDIT
	msg("%s", CREDIT);
#endif
#ifdef CREDIT2
	msg("%s", CREDIT2);
#endif
#ifdef COMPILED_BY
	msg("Compiled by %s", COMPILED_BY);
#endif
#ifdef COPYING
	msg("%s", COPYING);
#endif
}


/*ARGSUSED*/
void cmd_comment(frommark, tomark, cmd, bang, extra)
	MARK	frommark;
	MARK	tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
}


#ifdef SIGTSTP
/* temporarily suspend elvis */
/*ARGSUSED*/
void cmd_suspend(frommark, tomark, cmd, bang, extra)
	MARK	frommark;
	MARK	tomark;
	CMD	cmd;
	int	bang;
	char	*extra;
{
	msg("UnderB won't let you do ^Z");
	return;
}
#endif
