/* input.c */

/* Author:
 *	Steve Kirkendall
 *	1500 SW Park #326
 *	Portland OR, 97201
 *	kirkenda@cs.pdx.edu
 */


/* This file contains the input() function, which implements vi's INPUT mode.
 * It also contains the code that supports digraphs.
 */

#include "config.h"
#include "ctype.h"
#include "vi.h"


/* This function allows the user to replace an existing (possibly zero-length)
 * chunk of text with typed-in text.  It returns the MARK of the last character
 * that the user typed in.
 */
MARK input(from, to, when, delta)
	MARK	from;	/* where to start inserting text */
	MARK	to;	/* extent of text to delete */
	int	when;	/* either WHEN_VIINP or WHEN_VIREP */
	int	delta;	/* 1 to take indent from lower line, -1 for upper, 0 for none */
{
	unsigned char	key[3];		/* key char followed by '\0' char */
	unsigned char	keyTemp[2];	/* Hangul 1st byte followed by '\0'*/
					/* for v_subst() with Hangul input */
	int	nByteInput;

	char	*build;	/* used in building a newline+indent string */
	char	*scan;	/* used while looking at the indent chars of a line */
	MARK	m;	/* some place in the text */

	key[2] = 0;
	keyTemp[1] = 0;

	/* if we're replacing text with new text, save the old stuff */
	/* (Alas, there is no easy way to save text for replace mode) */
	if (from != to)
	{
		cut(from, to);
	}

	/* if doing a dot command, then reuse the previous text */
	if (doingdot)
	{
		ChangeText
		{
			/* delete the text that's there now */
			if (from != to)
			{
				delete(from, to);
			}

			/* insert the previous text */
			cutname('.');
			cursor = paste(from, FALSE, TRUE) + 1L;
		}
	}
	else /* interactive version */
	{
		int addspace = 0;
		/* assume that whoever called this already did a beforedo() */

		/* if doing a change within the line... */
		if (from != to && markline(from) == markline(to))
		{
			/* mark the end of the text with a "$" */
			pfetch(markline(to-1));
			if (IsHIdx(ptext, markidx(to-1)) == HAN_SECOND)
			{
				change(to - 2, to, " $");
			}
			else
			{
				change(to - 1, to, "$");
			}
		}
		else
		{
			/* delete the old text right off */
			if (from != to)
			{
				delete(from, to);
			}
			to = from;
		}

		/* handle autoindent of the first line, maybe */
		cursor = from;
		m = cursor + MARK_AT_LINE(delta);
		if (delta != 0 && *o_autoindent && markidx(m) == 0
		 && markline(m) >= 1L && markline(m) <= nlines)
		{
			/* Only autoindent blank lines. */
			pfetch(markline(cursor));
			if (plen == 0)
			{
				/* Okay, we really want to autoindent */
				pfetch(markline(m));
				for (scan = ptext, build = tmpblk.c;
				     *scan == ' ' || *scan == '\t';
				     )
				{
					*build++ = *scan++;
				}
				if (build > tmpblk.c)
				{
					*build = '\0';
					add(cursor, tmpblk.c);
					cursor += (int)(build - tmpblk.c);
					if (cursor > to)
						to = cursor;
				}
			}
		}

		/* repeatedly add characters from the user */
		nByteInput = 0;
		for (;;)
		{
			/* Get a character */
			redraw(cursor, TRUE);
			key[0] = getkey(when);
			if (IsHiBitOn( key[0] ))
			{
				key[1] = getkey(when);
				nByteInput = 2;
			}
			else
			{
				key[1] = 0;
				nByteInput = 1;
			}
#ifndef NO_VISIBLE
			if (key[0] != ctrl('O') && V_from != MARK_UNSET)
			{
				msg("Can't modify text during a selection");
				beep();
				continue;
			}
#endif

#ifndef CRUNCH
			/* if wrapmargin is set & we're past the
			 * warpmargin, then change the last whitespace
			 * characters on line into a newline
			 */
			if (*o_wrapmargin)
			{
				pfetch(markline(cursor));
				if (!ptext[markidx(cursor)]
				 && idx2col(cursor, ptext, TRUE) > COLS - (*o_wrapmargin & 0xff))
				{
					build = tmpblk.c;
					*build++ = '\n';
					if (*o_autoindent)
					{
						/* figure out indent for next line */
						for (scan = ptext; *scan == ' ' || *scan == '\t'; )
						{
							*build++ = *scan++;
						}
					}
					*build = '\0';

					scan = ptext + plen;
					m = cursor & ~(BLKSIZE - 1);
					while (ptext < scan)
					{
						scan--;
						if (*scan != ' ' && *scan != '\t')
							continue;

						/*break up line, and we do autoindent if needed*/
						change(m + (int)(scan - ptext), m + (int)(scan - ptext) + 1, tmpblk.c);

						/* NOTE: for some reason, MSC 5.10 doesn't
						 * like for these lines to be combined!!!
						 */
						cursor = (cursor & ~(BLKSIZE - 1));
						cursor += BLKSIZE;
						cursor += strlen(tmpblk.c) - 1;
						cursor += plen - (int)(scan - ptext) - 1;

						/*remove trailing spaces on previous line*/
						pfetch(markline(m));
						scan = ptext + plen;
						while (ptext < scan)
						{
							scan--;
							if (*scan != ' ' && *scan != '\t')
								break;
						}
						delete(m + (int)(scan - ptext) + 1, m + plen);

						break;
					}
				}
			}
#endif /* !CRUNCH */

			/* process it */
			switch (*key)
			{
			  case ctrl('['):
				/* if last line contains only whitespace, then remove whitespace */
				if (*o_autoindent)
				{
					pfetch(markline(cursor));
					for (scan = ptext; isspace(*scan); scan++)
					{
					}
					if (scan > ptext && !*scan)
					{
						cursor &= ~(BLKSIZE - 1L);
						if (to < cursor + plen)
						{
							to = cursor + plen;
						}
					}
				}
				goto BreakBreak;

			  case ctrl('U'):
				if (markline(cursor) == markline(from))
				{
					cursor = from;
				}
				else
				{
					cursor &= ~(BLKSIZE - 1);
				}
				break;

			  case ctrl('D'):
			  case ctrl('T'):
				if (to > cursor)
				{
					delete(cursor, to);
				}
				mark[27] = cursor;
				cmd_shift(cursor, cursor, *key == ctrl('D') ? CMD_SHIFTL : CMD_SHIFTR, TRUE, "");
				if (mark[27])
				{
					cursor = mark[27];
				}
				else
				{
					cursor = m_front(cursor, 0L);
				}
				to = cursor;
				break;

			  case '\b':
				if (cursor <= from)
				{
					beep();
				}
				else if (markidx(cursor) == 0)
				{
					cursor -= BLKSIZE;
					pfetch(markline(cursor));
					cursor += plen;
				}
				else
				{
					cursor = m_left(cursor, 1L);
				}
				break;

			  case ctrl('W'):
				m = m_bword(cursor, 1L, 'b');
				if (markline(m) == markline(cursor) && m >= from)
				{
					cursor = m;
					if (from > cursor)
					{
						from = cursor;
					}
				}
				else
				{
					beep();
				}
				break;

			  case '\n':
			  case '\r':
				build = tmpblk.c;
				*build++ = '\n';
				if (*o_autoindent)
				{
					/* figure out indent for next line */
					pfetch(markline(cursor));
					for (scan = ptext; *scan == ' ' || *scan == '\t'; )
					{
						*build++ = *scan++;
					}

					/* remove indent from this line, if blank */
					if ((int)(scan - ptext) == markidx(cursor) && plen > 0)
					{
						to = cursor &= ~(BLKSIZE - 1);
						delete(cursor, cursor + (int)(scan - ptext));
					}
				}
				*build = 0;
				if (cursor >= to && when != WHEN_VIREP)
				{
					add(cursor, tmpblk.c);
				}
				else
				{
					change(cursor, to, tmpblk.c);
				}
				redraw(cursor, TRUE);
				to = cursor = (cursor & ~(BLKSIZE - 1))
						+ BLKSIZE
						+ (int)(build - tmpblk.c) - 1;
				break;

			  case ctrl('A'):
			  case ctrl('P'):
				if (cursor < to)
				{
					delete(cursor, to);
				}
				if (*key == ctrl('A'))
				{
					cutname('.');
				}
				m = paste(cursor, FALSE, TRUE);
				if (m != MARK_UNSET)
				{
					to = cursor = m + 1L;
				}
				break;

			  case ctrl('V'):
				if (cursor >= to && when != WHEN_VIREP)
				{
					add(cursor, "^");
				}
				else
				{
					change(cursor, to, "^");
					to = cursor + 1;
				}
				redraw(cursor, TRUE);
				*key = getkey(0);
				if (*key == '\n')
				{
					/* '\n' too hard to handle */
					*key = '\r';
				}
				change(cursor, cursor + 1, key);
				cursor++;
				if (cursor > to)
				{
					to = cursor;
				}
				break;

			  case ctrl('L'):
			  case ctrl('R'):
				redraw(MARK_UNSET, FALSE);
				break;

			  default:
				if (cursor >= to && when != WHEN_VIREP)
				{
					add(cursor, key);
					cursor += nByteInput;
					to = cursor;
				}
				else
				{
					pfetch(markline(cursor));
					if (markidx(cursor) == plen)
					{
						add(cursor, key);
						addspace = 0;
					}
					else
					{
						if (cursor + nByteInput < to || when == WHEN_VIREP)
						{
							if ((nByteInput == 1) && IsHiBitOnMark(cursor))
							{ /* An ASC over 1st half of HANGUL */
								change(cursor+1, cursor+2, " ");
								addspace = 1;
							}
							else if ((nByteInput==2) && !IsHiBitOnMark(cursor) && IsHiBitOnMark(cursor+1))
							{ /* an HANGUL over ASC+HANGUL */
								change(cursor+2, cursor+3, " ");
								addspace = 1;
							}
							else
							{
								addspace = 0;
							}
							change(cursor, cursor + nByteInput, key);
						}
						else if (cursor + nByteInput <= to )
						{
							change(cursor, cursor+nByteInput, key);
							addspace = 0;
						}
						else
						{ /* nByteInput crosses to */
							keyTemp[0] = key[0];
							change(cursor, cursor+1, keyTemp);
							if (nByteInput == 2)
							{ /* add 2nd byte */
								add(cursor+1, &key[1]);
								to++;
							}
							addspace = 0;
						}
					}
					cursor += nByteInput;
				}
#ifndef NO_SHOWMATCH
				/* show matching "({[" if necessary */
				if (*o_showmatch && strchr(")}]", *key))
				{
					redraw(cursor, TRUE);
					m = m_match(cursor - 1, 0L);
					if (markline(m) >= topline
					 && markline(m) <= botline)
					{
						redraw(m, TRUE);
						refresh();
						sleep(1);
					}
				}
#endif
			} /* end switch(*key) */
		} /* end for(;;) */
BreakBreak:;
		/* delete any excess characters */
		if (cursor < to)
		{
			delete(cursor, to);
		}
		else if ((when == WHEN_VIREP) && addspace)
		{
			delete(cursor, cursor+1);
		}
	} /* end if doingdot else */

	/* put the new text into a cut buffer for possible reuse */
	if (!doingdot)
	{
		blksync();
		cutname('.');
		cut(from, cursor);
	}

	/* move to last char that we inputted, unless it was newline */
	if (markidx(cursor) != 0)
	{
		cursor--;
		pfetch(markline(cursor));
		if (IsHIdx(ptext,markidx(cursor)) == HAN_SECOND)
			cursor--;
	}
	redraw(cursor, FALSE);
	rptlines = 0L;
	return cursor;
}
